/*
	ProWesS, PROGS Windowing System

	File Select type "FSEL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 5 September 1995
	26.10.1995 added a user usable box below outline + access to outline
	28.10.1995 added queries for menu or directory in window (for user changed
			   fileselect windows) + made sure the global auxiliary can be used
			   by the client who extends the fileselect window
			   window had a sleep item - which was quite wrong
			   added option whether subdirectories should be displayed or not
	26.11.1995 "quit" now only quits, and a "do" item added
	18.05.1996 added PW_FILESELECT_FILENAME
	27.06.1996 DO now only works when a file has been selected (if not multiple)
*/

#include "mem_h"
#include "dir_h"
#include "str_h"
#include "dev_h"
#include "ctype_h"

#include "PWHandler_h"

#define TYPE PW_TYPE_FILESELECT
#define catch(X) if (err=(X)) return err; else

/* loose item object definition */

typedef struct {
	PWBaseObject;
	PWObject outl, box;
	PWObject dir, name, ext;
	PWObject all, tree, not;
	PWObject files, dirmenu;
	Error (*After)(PWObject obj, char *dir, void *file);
	char multiple;					/* TRUE if multiple files can be selected */
	char showsub;					/* TRUE when subdirs should be displayed */
	char do_exit;					/* TRUE when exitted with a DO */
	} Object;

/* forward declarations - for file select */
#define EXTENSIONLEN 128
Error getdir(PWObject object);
Error dirup(PWObject object);
Error fsel(PWObject object, char *item);
Error exit(PWObject object);

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;

	/* allocate specific object - which should be cleared (automatic) */
	catch( PWAllocate(event.region->system,sizeof(Object),ret) );

	((Object *)(*ret))->showsub=TRUE;

	/* see if there is a PW_EDLINE_MAXLENGTH tag */
	while (*taglist)
	{
		if (*taglist==PW_FILESELECT_MULTIPLE)
		{
			((Object *)(*ret))->multiple=TRUE;
			*taglist&=PW_TAGMASK_SKIP;
		}
		taglist=PWSkipTag(taglist);
	}
	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;

	if (obj->outl)
		PWRemove(obj->outl);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error Query(PWObject object, int what, void *ret)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;

	switch (what)
	{
		case PW_FILESELECT_BOX:
			*((PWObject *)ret)=obj->box;
			break;
		case PW_FILESELECT_MENU:
			*((PWObject *)ret)=obj->files;
			break;
		case PW_FILESELECT_OUTLINE:
			*((PWObject *)ret)=obj->outl;
			break;
		case PW_FILESELECT_DIRECTORY:
			err=PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,ret,NULL);
			break;
		case PW_FILESELECT_FILENAME:
			err=PWChange(obj->name,PW_EDLINE_GET,IO_MAXFULLNAME,ret,NULL);
			break;
		default:
			err=ERR_IPAR;
	}
	return err;
}


/* actions for file select window */

Error getdir(PWObject object)
{
	Object *obj;
	DirIterator it;
	char currdir[IO_MAXDIRECTORY];
	char name[IO_MAXDIRECTORY+2];
	char ext[EXTENSIONLEN];
	int not,tree;
	FileInfo fi;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);
	if (*currdir=='\0')
		DEVDataGet(currdir);
	PWChange(obj->dirmenu,PW_DIRSELECT_SET,currdir,NULL);
	PWChange(obj->files,PW_MENU_CLEAR,NULL);
	PWChange(obj->ext,PW_EDLINE_GET,EXTENSIONLEN,ext,NULL);
	PWQuery(obj->not,PW_LOOSE_STATUS,&not);
	PWQuery(obj->tree,PW_LOOSE_STATUS,&tree);
	not=(not==PW_STATUS_SELECTED);
	tree=(tree==PW_STATUS_SELECTED);

	/* get subdirectories */
	if (obj->showsub && !tree && !DIRSubOpen(currdir,FALSE,&it))
	{
		while(!DIRSubNext(it,&fi))
		{
			STRCopy("> ",name);
			STRAppend(fi.name,name);
			PWChange(obj->files,PW_MENU_ADD_COPY,name,NULL);
		}
		DIRSubClose(it);
	}

	/* get files in directory */
	if (!DIROpen(currdir,tree,&it,
				 not?DIRSieveExtensionNot:DIRSieveExtension,ext,NULL))
	{
		DIRDirectory(it,currdir);
		PWChange(obj->dir,PW_EDLINE_SET,currdir,NULL);
		while(!DIRNext(it,&fi))
			PWChange(obj->files,PW_MENU_ADD_COPY,fi.name,NULL);
		DIRClose(it);
	}

	/* select files if ALL is selected */
	if (obj->multiple)
	{
		int status;
		char *item;
		PWQuery(obj->all,PW_LOOSE_STATUS,&status);
		if (status==PW_STATUS_SELECTED)
		{
			PWQuery(obj->files,PW_MENU_FIRST,&item);
			while (item)
			{
				if (*item!='>')
					PWChange(obj->files,
								PW_MENU_STATUS_CURRENT,PW_STATUS_SELECTED,
								NULL);
				PWQuery(obj->files,PW_MENU_NEXT,&item);
			}
		}
	}

	return ERR_OK;
}

Error dirup(PWObject object)
{
	Object *obj;
	char currdir[IO_MAXDIRECTORY];
	int len;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);

	len=strlen(currdir);
	for (len=strlen(currdir)-1 ; len>0 && currdir[len-1]!='_' ; len--);
	if (len>0)
	{
		currdir[len]='\0';
		PWChange(obj->dir,PW_EDLINE_SET,currdir,NULL);
		getdir(object);
	}
	return ERR_OK;
}

Error dirdown(PWObject object, Object *obj, char *item)
{
	char currdir[IO_MAXDIRECTORY];

	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);

	if (currdir[strlen(currdir)-1]!='_')
		STRAppendSafe("_",currdir,IO_MAXDIRECTORY);
	STRAppendSafe(item,currdir,IO_MAXDIRECTORY);
	PWChange(obj->dir,PW_EDLINE_SET,currdir,NULL);
	getdir(object);
	object->event.region->system->ptr_info.keystroke=(char)-1;

	return ERR_OK;
}

Error fsel(PWObject object, char *item)
{
	Object *obj;
	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	if (*item=='>')
		return dirdown(object, obj,item+2);
	else
	{
		PWChange(obj->name,PW_EDLINE_SET,item,NULL);
		return ERR_OK;
	}
}

Error all(PWObject object)
{
	Object *obj;
	int status;
	char *item;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	PWQuery(obj->all,PW_LOOSE_STATUS,&status);

	/* select files if ALL is selected */
	if (status!=PW_STATUS_UNAVAILABLE)
	{
		PWQuery(obj->files,PW_MENU_FIRST,&item);
		while (item)
		{
			if (*item!='>')
				PWChange(obj->files,PW_MENU_STATUS_CURRENT,status,NULL);
			PWQuery(obj->files,PW_MENU_NEXT,&item);
		}
	}
	return ERR_OK;
}

/* actions for directory select window */

Error dirselect(PWObject object)
{
	Object *obj;
	char currdir[IO_MAXDIRECTORY];

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	return PWChange(obj->dirmenu,PW_DIRSELECT_ACTIVATE,NULL);
}

Error dirchange(PWObject object, char *currdir)
{
	Object *obj;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	PWChange(obj->dir,PW_EDLINE_SET,currdir,NULL);
	return getdir(object);
}

Error Handle(PWObject object, PWWait event)
{
	Object *obj=(Object *)object;
	char dir[IO_MAXDIRECTORY];
	char name[IO_MAXFULLNAME];
	Error err;

	obj->do_exit=FALSE;
	getdir(obj->outl);
	PWChange(obj->outl,PW_WINDOW_ORIGIN_RESET,NULL);
	catch( PWActivate(obj->outl) );
	PWChange(obj->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,dir,NULL);
	if (!obj->multiple)
		PWChange(obj->name,PW_EDLINE_GET,IO_MAXFULLNAME,name,NULL);
	if (obj->do_exit)
	{
		if (obj->After)
		{
			if (obj->multiple)
				err=obj->After((PWObject)obj,dir,obj->files);
			else
				err=obj->After((PWObject)obj,dir,name);
		}
	}
	return err;
}

Error exit(PWObject object)
{
	Object *obj;
	char name[2];

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	if (obj->multiple)
		name[0]=1;
	else
		PWChange(obj->name,PW_EDLINE_GET,2,name,NULL);
	if (name[0])
	{
		obj->do_exit=TRUE;
		PWChange(object,PW_SYSTEM_BREAKDOWN,NULL);
	}
	return ERR_OK;
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		PWObject dirbox, namebox, extbox;

		obj->status|=OSTATUS_INITIALISED;

		/* create main window */
		catch( PWCreate(NULL,&obj->outl,PW_TYPE_OUTLINE,
						PW_OUTLINE_QUIT,
						PW_OUTLINE_ACTION_WAKE, getdir,
						PW_OUTLINE_QUIT_KEYPRESS, 27,
						PW_OUTLINE_TITLE_TEXT, "file select",
						PW_OUTLINE_ACTION_DO, exit,
						PW_WINDOW_YSIZE, short2pt(540),
						PW_OBJECT_AUXILIARY, obj,
						NULL) );

		/* create user fillable box */
		catch( PWCreate(obj->outl,&obj->box,PW_TYPE_DIRECTION,
						PW_SCALE_FACTOR, 0,
						NULL) );

		/* create edline for filename, with label */
		catch( PWCreate(obj->outl,&namebox,PW_TYPE_DIRECTION,
						PW_SCALE_FACTOR, 0,
						NULL) );
		if (!obj->multiple)
		{
			catch( PWCreate(namebox,NULL,PW_TYPE_LABEL,
							PW_LABEL_TEXT, "filename",
							PW_SCALE_FACTOR, 0,
							NULL) );
			catch( PWCreate(namebox,&obj->name,PW_TYPE_EDLINE,
							PW_EDLINE_MAXLENGTH, IO_MAXFULLNAME,
							PW_EDLINE_KEYPRESS, 'f',
							PW_EDLINE_ACTION_AFTER, exit,
							NULL) );
		}
		/* create edline for extensions, with label */
		catch( PWCreate(obj->outl,&extbox,PW_TYPE_DIRECTION,
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(extbox,NULL,PW_TYPE_LABEL,
						PW_LABEL_TEXT, "extensions",
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(extbox,&obj->ext,PW_TYPE_EDLINE,
						PW_EDLINE_MAXLENGTH, EXTENSIONLEN,
						PW_EDLINE_KEYPRESS, 'e',
						PW_EDLINE_WIDTH_FS, short2pt(12),
						PW_EDLINE_ACTION_AFTER, getdir,
						NULL) );
		catch( PWCreate(extbox,&obj->not,PW_TYPE_LOOSE_ITEM,
						PW_LOOSE_TEXT, "Not",
						PW_LOOSE_KEYPRESS, 'n',
						PW_SCALE_FACTOR, 0,
						PW_LOOSE_ACTION_HIT, getdir,
						NULL) );

		/* create edline for directory, with label */
		catch( PWCreate(obj->outl,&dirbox,PW_TYPE_DIRECTION,
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(dirbox,NULL,PW_TYPE_LABEL,
						PW_LABEL_TEXT, "directory",
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(dirbox,&obj->dir,PW_TYPE_EDLINE,
						PW_EDLINE_MAXLENGTH, IO_MAXDIRECTORY,
						PW_EDLINE_KEYPRESS, 'd',
						PW_EDLINE_ACTION_AFTER, getdir,
						PW_EDLINE_WIDTH_FS, short2pt(12),
						PW_EDLINE_ACTION_DO, dirselect,
						NULL) );
		catch( PWCreate(dirbox,&obj->dirmenu,PW_TYPE_DIRSELECT,
						PW_KEYPRESS, 'D',
						PW_DIRSELECT_ACTION, dirchange,
						NULL) );
		catch( PWCreate(dirbox,NULL,PW_TYPE_LOOSE_ITEM,
						PW_LOOSE_CHANGE_STATUS,FALSE,
						PW_LOOSE_ACTION_HIT, dirup,
						PW_LOOSE_TEXT, "<-",
						PW_LOOSE_KEYPRESS, '<',
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(dirbox,&obj->tree,PW_TYPE_LOOSE_ITEM,
						PW_LOOSE_TEXT, "Tree",
						PW_LOOSE_KEYPRESS, 't',
						PW_SCALE_FACTOR, 0,
						PW_LOOSE_ACTION_HIT, getdir,
						NULL) );
		if (obj->multiple)
		{
			catch( PWCreate(dirbox,&obj->all,PW_TYPE_LOOSE_ITEM,
							PW_LOOSE_TEXT, "All",
							PW_LOOSE_KEYPRESS, 'a',
							PW_LOOSE_ACTION_HIT, all,
							PW_SCALE_FACTOR, 0,
							NULL) );
		}
		/* create menu with possible filenames */
		catch( PWCreate(obj->outl,NULL,PW_TYPE_SEPARATOR,NULL) );
		catch( PWCreate(obj->outl,&obj->files,PW_TYPE_MENU,
						PW_MENU_VISIBLE_LINES, 4,
						PW_MENU_ACTION_SELECT, fsel,
						PW_MENU_SORT_COMPARE, STRCompareCI,
						PW_MENU_WINDOW_DO, TRUE,
						NULL) );
		if (!obj->multiple)
			PWChange(obj->files,PW_MENU_NONE_SELECTED,NULL);
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_FILESELECT_TITLE_TEXT:
			PWChange(obj->outl,PW_OUTLINE_TITLE_TEXT,(char *)(*(taglist+1)),NULL);
			break;
		case PW_FILESELECT_ACTIVATE:
			Handle(object,0);
			break;
		case PW_FILESELECT_ACTION:
			obj->After=(Error (*)(PWObject,char *,void *))(*(taglist+1));
			break;
		case PW_FILESELECT_FILENAME:
			if (obj->name)
				PWChange(obj->name,PW_EDLINE_SET,(*(taglist+1)),NULL);
			break;
		case PW_FILESELECT_DIRECTORY:
			PWChange(obj->dir,PW_EDLINE_SET,(*(taglist+1)),NULL);
			break;
		case PW_FILESELECT_EXTENSION:
			PWChange(obj->ext,PW_EDLINE_SET,(*(taglist+1)),NULL);
			break;
		case PW_FILESELECT_NOT_STATUS:
			PWChange(obj->not,PW_LOOSE_STATUS,(*(taglist+1)),NULL);
			break;
		case PW_FILESELECT_SHOWSUB:
			obj->showsub=(*(taglist+1));
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, TYPE, PW_TYPE_EVENT_KEYPRESS,
	&Create, &Change, &Query, &Remove, NULL, NULL, &Handle, NULL,
	NULL, NULL, NULL
	};
