/*
	ProWesS, PROGS Windowing System

	Scroll bar object "SCRO"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started May 2 1995
	22.03.1996 adjusted for internal use of pixel values in ProWesS
	27.03.1996 added support for do/hit on scroll bar
			   HIT will move to the indicated position
			   DO will move to start/end of doc, depending on half where hit
	10.04.1996 use PWHandleDefine && use ProWesS default colours
	11.05.1996 PW_SCROLL_MIN/MAX_KEYPRESSES added
	06.06.1996 the PW_SCROLL_MIN/MAX_KEYPRESSES should now also work when the
			   object auxiliary was set
	11.06.1996 PW_LOOSE_AUTOREPEAT on scroll items
	12.07.1996 use PFP variants for drawing scroll bar
*/

#include "mem_h"

#include "ProWesS_h"
#include "PWTypeDef_h"
#include "PWHandler_h"

#define catch(X) if (err=(X)) return err; else

#define p100 short2pt(100)
pix xwidth=2, ywidth=2; 			/* width nominal and in each direction */
pt xarrow=6, yarrow=6; pt width;	/* size nominal and in each direction */
ColourRGB colour = { p100,    0,    0 }; char Kcolour =FALSE;
ColourRGB acolour= { p100,    0,    0 }; char Kacolour=FALSE;
ColourRGB back	 = { p100, p100, p100 }; char Kback   =FALSE;
char arrowsleft=FALSE;
char arrowsabove=FALSE;

/* global statusses */

/* object definition */
typedef struct _Object {
	PWBaseObject;					/* common start for all objects */
	PWObject canvas;				/* canvas connected with this bar */
	PWObject upleft, downright, bar;/* objects inside this one */
	pt mindist, maxdist;			/* distance for scroll on hit/do */
	pt minimum, maximum, position;	/* limits and current position */
	char nobar; 					/* TRUE when we only want arrows */
	Error (*calcsize)(struct _Object *, pt *);	/* calculate canvas' size */
	PWObject key1,key2,key3,key4;	/* possible keyspresses for scrolling */
} Object;

/* object handling routines */

Error ScrollCreate(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;
	int skip;

	LinkPROforma();
	/* allocate specific object - which should be cleared (automatic) */
	catch( PWAllocate(event.region->system,sizeof(Object),ret) );

	if (!Kcolour) PWQuery(*ret,PW_DEFAULT_MIDDLEGROUND,&colour);
	if (!Kacolour) PWQuery(*ret,PW_DEFAULT_FOREGROUND,&acolour);
	if (!Kback) PWQuery(*ret,PW_DEFAULT_BACKGROUND,&back);

	/* no events to react upon - all done by children */
	event.region->scale=0;			/* don't scale a scrollbar */
	event.region->composite=TRUE;

	/* see if there is a PW_SCROLL_NOBAR tag */
	while (*taglist)
	{
		if (*taglist==PW_SCROLL_NOBAR)
		{
			((Object *)(*ret))->nobar=TRUE;
			*taglist&=PW_TAGMASK_SKIP;
		}
		taglist=PWSkipTag(taglist);
	}
	return err;
}

Error ScrollRemove(PWObject object)
{
	Object *obj=(Object *)object;

	PWRemove(obj->key1); PWRemove(obj->key2);
	PWRemove(obj->key3); PWRemove(obj->key4);

	/* all the children are automatically removed */
	return PWRelease((PWObject)obj,obj);
}

Error SetTRUE(Defines *this, char *value)
{
	*(char *)this->data=TRUE;
	return ERR_OK;
}
Error SetFALSE(Defines *this, char *value)
{
	*(char *)this->data=FALSE;
	return ERR_OK;
}
void cwidth(void *data)
{
	xwidth=PF2xpix(width);
	ywidth=PF2ypix(width);
}
void carrow(void *data)
{
	xarrow=PFxround(width);
	yarrow=PFyround(width);
}

Defines defines[]={
{ "ARROWS-LEFT",            SetTRUE,            NULL,   &arrowsleft,    NULL },
{ "ARROWS-RIGHT",           SetFALSE,           NULL,   &arrowsleft,    NULL },
{ "ARROWS-ABOVE",           SetTRUE,            NULL,   &arrowsabove,   NULL },
{ "ARROWS-BELOW",           SetFALSE,           NULL,   &arrowsabove,   NULL },
{ "BAR-MARGINWIDTH",        PW_DEFINE_PT,       cwidth, &width,         NULL },
{ "ARROW-SIZE",             PW_DEFINE_PT,       carrow, &width,         NULL },
{ "BAR-BACKGROUND-COLOUR",  PW_DEFINE_COLOUR,   NULL,   &back,          &Kback },
{ "BAR-COLOUR",             PW_DEFINE_COLOUR,   NULL,   &colour,        &Kcolour },
{ "ARROW-COLOUR",           PW_DEFINE_COLOUR,   NULL,   &acolour,       &Kacolour },
{ 0 }
};

Error ScrollDefine(char *name, char *value)
{
	return PWHandleDefine("SCROLL-",defines,name,value);
}

static Error drawtrileft(PWObject object, Gstate gstate, pt xsiz, pt ysiz)
{
	pt xorg=(xsiz-xarrow)/2;
	pt yorg=(ysiz-yarrow)/2;

	PFPathFilled(gstate);
	PFColourRGB(gstate, &acolour);
	PFMoveTo(gstate, xorg, yorg+yarrow/2);
	PFLineTo(gstate, xorg+xarrow, yorg);
	PFLineR(gstate, 0, yarrow);
	PFPathClose(gstate);
	PFPathDraw(gstate);
}

static Error drawtriright(PWObject object, Gstate gstate, pt xsiz, pt ysiz)
{
	pt xorg=(xsiz-xarrow)/2;
	pt yorg=(ysiz-yarrow)/2;

	PFPathFilled(gstate);
	PFColourRGB(gstate, &acolour);
	PFMoveTo(gstate, xorg, yorg);
	PFLineR(gstate, 0, yarrow);
	PFLineTo(gstate, xorg+xarrow, yorg+yarrow/2);
	PFPathClose(gstate);
	PFPathDraw(gstate);
}

static Error drawtriup(PWObject object, Gstate gstate, pt xsiz, pt ysiz)
{
	pt xorg=(xsiz-xarrow)/2;
	pt yorg=(ysiz-yarrow)/2;

	PFPathFilled(gstate);
	PFColourRGB(gstate, &acolour);
	PFMoveTo(gstate, xorg, yorg+yarrow);
	PFLineR(gstate, xarrow, 0);
	PFLineTo(gstate, xorg+xarrow/2, yorg);
	PFPathClose(gstate);
	PFPathDraw(gstate);
}

static Error drawtridown(PWObject object, Gstate gstate, pt xsiz, pt ysiz)
{
	pt xorg=(xsiz-xarrow)/2;
	pt yorg=(ysiz-yarrow)/2;

	PFPathFilled(gstate);
	PFColourRGB(gstate, &acolour);
	PFMoveTo(gstate, xorg, yorg);
	PFLineR(gstate, xarrow, 0);
	PFLineTo(gstate, xorg+xarrow/2, yorg+yarrow);
	PFPathClose(gstate);
	PFPathDraw(gstate);
}

static Error bar_draw(PWObject object, CanvasInfo *rinfo)
{
	int scrolldirection, sizedirection, posdirection;
	pt *changesize, *changepos, minsize;
	Gstate gstate=rinfo->gstate;
	pt size, pos;
	double totsize;
	pt xpos, ypos, xsiz, ysiz;
	Error err=ERR_OK;
	Object *obj;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	PFPaperColourRGB(gstate, &back);
	PFPageClear(gstate);

	if (obj->canvas==NULL) return err;

	xpos=short2pt(xwidth); ypos=short2pt(ywidth);
	xsiz=short2pt(rinfo->xpixsiz-2*xwidth);
	ysiz=short2pt(rinfo->ypixsiz-2*ywidth);

	/* determine size and position of/in canvas */
	if (obj->event.region->level&1)
	{
		changesize=&ysiz;
		changepos=&ypos;
		minsize=ypix2PF(1);
		sizedirection=PW_CANVAS_YSIZE;
		posdirection=PW_CANVAS_YORIGIN;
	}
	else
	{
		changesize=&xsiz;
		changepos=&xpos;
		minsize=xpix2PF(1);
		sizedirection=PW_CANVAS_XSIZE;
		posdirection=PW_CANVAS_XORIGIN;
	}
	PWQuery(obj->canvas,sizedirection,&size);
	if (obj->calcsize) catch( obj->calcsize(obj,&size) );
	PWQuery(obj->canvas,posdirection,&pos);
	totsize=obj->maximum-obj->minimum;

	if (totsize>0 && size<totsize)
	{
		double ratio=(double)*changesize/totsize;
		pt end=*changepos+*changesize;

		*changepos=*changepos+(pos-obj->minimum)*ratio;
		*changesize=size*ratio;
		if (*changesize<minsize) *changesize=minsize;
		if ((*changepos+*changesize)>end) *changesize=end-*changepos;
	}

	PFPathFilled(gstate);
	PFColourRGB(gstate, &colour);
	PFPMoveTo(gstate, xpos, ypos);
	PFPLineR(gstate, xsiz, 0);
	PFPLineR(gstate, 0, ysiz);
	PFPLineR(gstate, -xsiz, 0);
	PFPathClose(gstate);
	PFPathDraw(gstate);

	return err;
}

static Error doscroll(Object *obj, char direction, pt amount)
{
	Error err=ERR_OK;
	int scrolldirection, sizedirection, posdirection;
	pt size, pos;

	if (obj->canvas==NULL) return err;

	if (obj->event.region->level&1)
	{
		scrolldirection=PW_CANVAS_YSCROLL;
		sizedirection=PW_CANVAS_YSIZE;
		posdirection=PW_CANVAS_YORIGIN;
	}
	else
	{
		scrolldirection=PW_CANVAS_XSCROLL;
		sizedirection=PW_CANVAS_XSIZE;
		posdirection=PW_CANVAS_XORIGIN;
	}

	/* determine size and position of/in canvas */
	PWQuery(obj->canvas,sizedirection,&size);
	if (obj->calcsize)
		catch( obj->calcsize(obj,&size) );
	PWQuery(obj->canvas,posdirection,&pos);

	/* allow scrolling of something less than a screen */
	if (amount<0)
	{
		if (size+amount>0)
			amount=size+amount;
		else
			amount=-amount;
	}

	/* make amount point in right direction */
	if (direction<0 )amount=-amount;

	/* make sure we don't scroll out of bounds */
	if (pos+amount+size>obj->maximum)
		amount=obj->maximum-size-pos;
	if (pos+amount<obj->minimum)
		amount=obj->minimum-pos;

	/* do the actual scrolling */
	if (amount!=0)
	{
		PWChange(obj->bar,PW_CANVAS_REDRAW,NULL);
		catch( PWChange(obj->canvas, scrolldirection, amount, NULL) );
	}
	return err;
}

static Error bar_hit(PWObject object, CanvasInfo *rinfo)
{
	int sizedirection, cposdirection;
	pt size, newpos, pos;
	pix pixsiz, pixpos;
	Object *obj;
	Error err;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	if (obj->canvas==NULL) return ERR_OK;

	/* determine size of canvas */
	if (obj->event.region->level&1)
	{
		sizedirection=PW_CANVAS_YSIZE;
		cposdirection=PW_CANVAS_YORIGIN;
		pixsiz=rinfo->ypixsiz;
		pixpos=rinfo->ypixpos-ywidth;
	}
	else
	{
		sizedirection=PW_CANVAS_XSIZE;
		cposdirection=PW_CANVAS_XORIGIN;
		pixsiz=rinfo->xpixsiz;
		pixpos=rinfo->xpixpos-xwidth;
	}
	PWQuery(obj->canvas,sizedirection,&size);
	PWQuery(obj->canvas,cposdirection,&pos);
	if (obj->calcsize) catch( obj->calcsize(obj,&size) );

	newpos=((double)(obj->maximum-obj->minimum))/pixsiz*pixpos - size/2.0;

	if (newpos==pos) return ERR_OK;
	if (newpos>pos)
		doscroll(obj, 1,newpos-pos);
	else
		doscroll(obj,-1,pos-newpos);

	return ERR_OK;
}

static Error bar_do(PWObject object, CanvasInfo *rinfo)
{
	int sizedirection, cposdirection;
	pt newpos, pos;
	pix pixsiz, pixpos;
	Object *obj;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	if (obj->canvas==NULL) return ERR_OK;

	/* determine size of canvas */
	if (obj->event.region->level&1)
	{
		cposdirection=PW_CANVAS_YORIGIN;
		if (rinfo->ypixpos>rinfo->ypixsiz/2)
			newpos=obj->maximum;
		else
			newpos=obj->minimum;
	}
	else
	{
		cposdirection=PW_CANVAS_XORIGIN;
		if (rinfo->xpixpos>rinfo->xpixsiz/2)
			newpos=obj->maximum;
		else
			newpos=obj->minimum;
	}

	PWQuery(obj->canvas,cposdirection,&pos);
	if (newpos==pos) return ERR_OK;
	if (newpos>pos)
		doscroll(obj, 1,newpos-pos);
	else
		doscroll(obj,-1,pos-newpos);

	return ERR_OK;
}

static Error scrollupleft(PWObject object)
{
	Object *obj;
	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	return doscroll(obj,-1,obj->mindist);
}

static Error scrolldownright(PWObject object)
{
	Object *obj;
	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	return doscroll(obj,1,obj->mindist);
}

static Error bigscrollupleft(PWObject object)
{
	Object *obj;
	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	return doscroll(obj,-1,obj->maxdist);
}

static Error bigscrolldownright(PWObject object)
{
	Object *obj;
	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	return doscroll(obj,1,obj->maxdist);
}

static Error arrows(Object *obj, int upleft, int downright)
{
	Error err;

	catch( PWCreate((PWObject)obj, &obj->upleft, PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_ACTION_DRAW, upleft,
					PW_LOOSE_SIZE, xarrow, yarrow,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_SCALE_FACTOR, obj->nobar?1:0,
					PW_LOOSE_ACTION_HIT, &scrollupleft,
					PW_LOOSE_ACTION_DO, &bigscrollupleft,
					PW_OBJECT_AUXILIARY, obj,
					PW_LOOSE_AUTOREPEAT,
					NULL) );
	catch( PWCreate((PWObject)obj, &obj->downright, PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_ACTION_DRAW, downright,
					PW_LOOSE_SIZE, xarrow, yarrow,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_SCALE_FACTOR, obj->nobar?1:0,
					PW_LOOSE_ACTION_HIT, &scrolldownright,
					PW_LOOSE_ACTION_DO, &bigscrolldownright,
					PW_OBJECT_AUXILIARY, obj,
					PW_LOOSE_AUTOREPEAT,
					NULL) );
}

Error ScrollChange(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		int upleft, downright;
		char *aboveleft;

		obj->status|=OSTATUS_INITIALISED;
		obj->auxiliary=obj;

		/* create arrows and bar objects */
		/* need to now the direction of this scrollbar */
		if (obj->event.region->level&1)
		{
			upleft=(int)drawtriup;
			downright=(int)drawtridown;
			aboveleft=&arrowsabove;
		}
		else
		{
			upleft=(int)drawtrileft;
			downright=(int)drawtriright;
			aboveleft=&arrowsleft;
		}
		/* horizontal */
		if (*aboveleft)
			catch( arrows(obj,upleft,downright) );
		if (obj->nobar==FALSE)
			catch( PWCreate((PWObject)obj,&obj->bar,PW_TYPE_CANVAS,
							PW_CANVAS_ACTION_REDRAW, bar_draw,
							PW_CANVAS_ACTION_HIT, bar_hit,
							PW_CANVAS_ACTION_DO, bar_do,
							PW_OBJECT_AUXILIARY, obj,
							NULL) );
		if (!*aboveleft)
			catch( arrows(obj,upleft,downright) );
	}

	while (*taglist)
	{
		int option=NUL; PWObject client=obj->canvas;

		switch (*taglist)
		{
		case PW_SCROLL_SCROLL:
			if ((*(taglist+1))>0)
				doscroll(obj, 1, (*(taglist+1)));
			else
				doscroll(obj,-1,-(*(taglist+1)));
			break;
		case PW_SCROLL_CALCSIZE:
			obj->calcsize=(Error (*)(Object *,pt*))(*(taglist+1));
			break;
		case PW_SCROLL_CANVAS:
			obj->canvas=(PWObject)(*(taglist+1));
			break;
		case PW_SCROLL_MINIMUM:
			obj->minimum=(pt)(*(taglist+1));
			PWChange(obj->bar,PW_CANVAS_REDRAW,NULL);
			break;
		case PW_SCROLL_MAXIMUM:
			obj->maximum=(pt)(*(taglist+1));
			PWChange(obj->bar,PW_CANVAS_REDRAW,NULL);
			break;
		case PW_SCROLL_MINDIST:
			obj->mindist=(pt)(*(taglist+1));
			if (obj->maxdist==0) obj->maxdist=obj->mindist;
			break;
		case PW_SCROLL_MAXDIST:
			obj->maxdist=(pt)(*(taglist+1));
			if (obj->mindist==0) obj->mindist=obj->maxdist;
			break;
		case PW_SCROLL_MIN_KEYPRESSES:
			if (!obj->key1)
				PWCreate(obj->upleft, &obj->key1, PW_TYPE_KEYPRESS,
						PW_KEYPRESS_ACTION, scrollupleft,
						NULL);
			if (!obj->key2)
				PWCreate(obj->upleft, &obj->key2, PW_TYPE_KEYPRESS,
						PW_KEYPRESS_ACTION, scrolldownright,
						NULL);
			PWChange(obj->key1,PW_KEYPRESS, *(taglist+1), NULL);
			PWChange(obj->key2,PW_KEYPRESS, *(taglist+2), NULL);
			break;
		case PW_SCROLL_MAX_KEYPRESSES:
			if (!obj->key3)
				PWCreate(obj->upleft, &obj->key3, PW_TYPE_KEYPRESS,
						PW_KEYPRESS_ACTION, bigscrollupleft,
						NULL);
			if (!obj->key4)
				PWCreate(obj->upleft, &obj->key4, PW_TYPE_KEYPRESS,
						PW_KEYPRESS_ACTION, bigscrolldownright,
						NULL);
			PWChange(obj->key3,PW_KEYPRESS, *(taglist+1), NULL);
			PWChange(obj->key4,PW_KEYPRESS, *(taglist+2), NULL);
			break;

		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_SCROLL, PW_TYPE_EVENT_REGION,
	&ScrollCreate, &ScrollChange, NULL, &ScrollRemove,
	NULL, NULL, NULL,
	&ScrollDefine, NULL, NULL, NULL
};
